@@ -23,10 +23,28 @@ from pyinfra.api import (
     operation,
     OperationError,
     OperationTypeError,
+    QuoteString,
     RsyncCommand,
+    StringCommand,
 )
-from pyinfra.api.connectors.util import escape_unix_path
-from pyinfra.api.util import get_file_sha1, get_template, memoize
+from pyinfra.api.command import make_formatted_string_command
+from pyinfra.api.util import (
+    get_file_sha1,
+    get_path_permissions_mode,
+    get_template,
+    memoize,
+)
+from pyinfra.facts.files import (
+    Directory,
+    File,
+    FindFiles,
+    FindInFile,
+    Link,
+    Md5File,
+    Sha1File,
+    Sha256File,
+)
+from pyinfra.facts.server import Date, Which
 
 from .util.compat import fspath
 from .util.files import (
@@ -73,9 +91,7 @@ def download(
         )
     '''
 
-    dest = escape_unix_path(dest)
-
-    info = host.fact.file(dest)
+    info = host.get_fact(File, path=dest)
     # Destination is a directory?
     if info is False:
         raise OperationError(
@@ -94,34 +110,42 @@ def download(
     else:
         if cache_time:
             # Time on files is not tz-aware, and will be the same tz as the server's time,
-            # so we can safely remove the tzinfo from host.fact.date before comparison.
-            cache_time = host.fact.date.replace(tzinfo=None) - timedelta(seconds=cache_time)
+            # so we can safely remove the tzinfo from the Date fact before comparison.
+            cache_time = host.get_fact(Date).replace(tzinfo=None) - timedelta(seconds=cache_time)
             if info['mtime'] and info['mtime'] > cache_time:
                 download = True
 
         if sha1sum:
-            if sha1sum != host.fact.sha1_file(dest):
+            if sha1sum != host.get_fact(Sha1File, path=dest):
                 download = True
 
         if sha256sum:
-            if sha256sum != host.fact.sha256_file(dest):
+            if sha256sum != host.get_fact(Sha256File, path=dest):
                 download = True
 
         if md5sum:
-            if md5sum != host.fact.md5_file(dest):
+            if md5sum != host.get_fact(Md5File, path=dest):
                 download = True
 
     # If we download, always do user/group/mode as SSH user may be different
     if download:
-        curl_command = "curl -sSLf '{0}' -o {1}".format(src, dest)
-        wget_command = "wget -q '{0}' -O {1} || (rm -f {1}; exit 1)".format(src, dest)
+        curl_command = make_formatted_string_command(
+            'curl -sSLf {0} -o {1}',
+            QuoteString(src),
+            QuoteString(dest),
+        )
+        wget_command = make_formatted_string_command(
+            'wget -q {0} -O {1} || ( rm -f {1} ; exit 1 )',
+            QuoteString(src),
+            QuoteString(dest),
+        )
 
-        if host.fact.which('curl'):
+        if host.get_fact(Which, command='curl'):
             yield curl_command
-        elif host.fact.which('wget'):
+        elif host.get_fact(Which, command='wget'):
             yield wget_command
         else:
-            yield '({0}) || ({1})'.format(curl_command, wget_command)
+            yield '( {0} ) || ( {1} )'.format(curl_command, wget_command)
 
         if user or group:
             yield chown(dest, user, group)
@@ -130,22 +154,22 @@ def download(
             yield chmod(dest, mode)
 
         if sha1sum:
-            yield (
-                '((sha1sum {0} 2> /dev/null || shasum {0} || sha1 {0}) | grep {1}) '
-                '|| (echo "SHA1 did not match!" && exit 1)'
-            ).format(dest, sha1sum)
+            yield make_formatted_string_command((
+                '(( sha1sum {0} 2> /dev/null || shasum {0} || sha1 {0} ) | grep {1} ) '
+                '|| ( echo {2} && exit 1 )'
+            ), QuoteString(dest), sha1sum, QuoteString('SHA1 did not match!'))
 
         if sha256sum:
-            yield (
-                '((sha256sum {0} 2> /dev/null || shasum -a 256 {0} || sha256 {0}) | grep {1}) '
-                '|| (echo "SHA256 did not match!" && exit 1)'
-            ).format(dest, sha256sum)
+            yield make_formatted_string_command((
+                '(( sha256sum {0} 2> /dev/null || shasum -a 256 {0} || sha256 {0} ) | grep {1}) '
+                '|| ( echo {2} && exit 1 )'
+            ), QuoteString(dest), sha256sum, QuoteString('SHA256 did not match!'))
 
         if md5sum:
-            yield (
-                '((md5sum {0} 2> /dev/null || md5 {0}) | grep {1}) '
-                '|| (echo "MD5 did not match!" && exit 1)'
-            ).format(dest, md5sum)
+            yield make_formatted_string_command((
+                '(( md5sum {0} 2> /dev/null || md5 {0} ) | grep {1}) '
+                '|| ( echo {2} && exit 1 )'
+            ), QuoteString(dest), md5sum, QuoteString('MD5 did not match!'))
 
     else:
         host.noop('file {0} has already been downloaded'.format(dest))
@@ -169,6 +193,7 @@ def line(
     + flags: list of flags to pass to sed when replacing/deleting
     + backup: whether to backup the file (see below)
     + interpolate_variables: whether to interpolate variables in ``replace``
+    + assume_present: whether to assume a matching line already exists in the file
 
     Regex line matching:
         Unless line matches a line (starts with ^, ends $), pyinfra will wrap it such that
@@ -235,15 +260,13 @@ def line(
         )
     '''
 
-    path = escape_unix_path(path)
-
     match_line = ensure_whole_line_match(line)
 
     # Is there a matching line in this file?
     if assume_present:
         present_lines = [line]
     else:
-        present_lines = host.fact.find_in_file(path, match_line)
+        present_lines = host.get_fact(FindInFile, path=path, pattern=match_line)
 
     # If replace present, use that over the matching line
     if replace:
@@ -253,15 +276,18 @@ def line(
         replace = ''
 
     # Save commands for re-use in dynamic script when file not present at fact stage
-    make_backup_command = ''
-    if backup:
-        make_backup_command = 'cp {0} {0}.{1} && '.format(path, get_timestamp())
-
-    echo_command = (
-        '{0}echo "{1}" >> {2}'.format(make_backup_command, line, path)
-        if interpolate_variables else
-        "{0}echo '{1}' >> {2}".format(make_backup_command, line, path)
+    echo_command = make_formatted_string_command(
+        'echo {0} >> {1}',
+        '"{0}"'.format(line) if interpolate_variables else QuoteString(line),
+        QuoteString(path),
     )
+
+    if backup:
+        backup_filename = '{0}.{1}'.format(path, get_timestamp())
+        echo_command = StringCommand(make_formatted_string_command(
+            'cp {0} {1} && ', QuoteString(path), QuoteString(backup_filename),
+        ), echo_command)
+
     sed_replace_command = sed_replace(
         path, match_line, replace,
         flags=flags,
@@ -271,25 +297,21 @@ def line(
 
     # No line and we want it, append it
     if not present_lines and present:
-        quoted_match_line = (
-            '"{0}"'.format(match_line)
-            if interpolate_variables else
-            "'{0}'".format(match_line)
-        )
-
         # If the file does not exist - it *might* be created, so we handle it
         # dynamically with a little script.
         if present_lines is None:
-            yield '''
-                if [ -f '{target}' ]; then
-                    (grep {quoted_match_line} '{target}' && {sed_replace_command}) 2> /dev/null || \
-                    {echo_command};
-                else
-                    {echo_command};
-                fi
-            '''.format(
-                target=path,
-                quoted_match_line=quoted_match_line,
+            yield make_formatted_string_command(
+                '''
+                    if [ -f '{target}' ]; then
+                        ( grep {match_line} '{target}' && \
+                        {sed_replace_command}) 2> /dev/null || \
+                        {echo_command} ;
+                    else
+                        {echo_command} ;
+                    fi
+                ''',
+                target=QuoteString(path),
+                match_line=QuoteString(match_line),
                 echo_command=echo_command,
                 sed_replace_command=sed_replace_command,
             )
@@ -299,8 +321,10 @@ def line(
             # If we're doing replacement, only append if the *replacement* line
             # does not exist (as we are appending the replacement).
             if replace:
-                present_lines = host.fact.find_in_file(
-                    path, ensure_whole_line_match(replace),
+                present_lines = host.get_fact(
+                    FindInFile,
+                    path=path,
+                    pattern=ensure_whole_line_match(replace),
                 )
 
             if not present_lines:
@@ -360,9 +384,7 @@ def replace(
         )
     '''
 
-    path = escape_unix_path(path)
-
-    existing_lines = host.fact.find_in_file(path, match)
+    existing_lines = host.get_fact(FindInFile, path=path, pattern=match)
 
     # Only do the replacement if the file does not exist (it may be created earlier)
     # or we have matching lines.
@@ -444,7 +466,7 @@ def sync(
             continue
 
         if remote_dirname:
-            ensure_dirnames.append(remote_dirname)
+            ensure_dirnames.append((remote_dirname, get_path_permissions_mode(dirname)))
 
         for filename in filenames:
             full_filename = os_path.join(dirname, filename)
@@ -467,21 +489,22 @@ def sync(
     # Ensure the destination directory - if the destination is a link, ensure
     # the link target is a directory.
     dest_to_ensure = dest
-    dest_link_info = host.fact.link(dest)
+    dest_link_info = host.get_fact(Link, path=dest)
     if dest_link_info:
         dest_to_ensure = dest_link_info['link_target']
 
     yield directory(
         dest_to_ensure,
         user=user, group=group,
+        mode=get_path_permissions_mode(src),
         state=state, host=host,
     )
 
     # Ensure any remote dirnames
-    for dirname in ensure_dirnames:
+    for dirname, dir_mode in ensure_dirnames:
         yield directory(
             '/'.join((dest, dirname)),
-            user=user, group=group,
+            user=user, group=group, mode=dir_mode,
             state=state, host=host,
         )
 
@@ -489,7 +512,8 @@ def sync(
     for local_filename, remote_filename in put_files:
         yield put(
             local_filename, remote_filename,
-            user=user, group=group, mode=mode,
+            user=user, group=group,
+            mode=mode or get_path_permissions_mode(local_filename),
             add_deploy_dir=False,
             create_remote_dir=False,  # handled above
             state=state, host=host,
@@ -497,7 +521,7 @@ def sync(
 
     # Delete any extra files
     if delete:
-        remote_filenames = set(host.fact.find_files(dest) or [])
+        remote_filenames = set(host.get_fact(FindFiles, path=dest) or [])
         wanted_filenames = set([remote_filename for _, remote_filename in put_files])
         files_to_delete = remote_filenames - wanted_filenames
         for filename in files_to_delete:
@@ -587,8 +611,6 @@ def get(
         )
     '''
 
-    src = escape_unix_path(src)
-
     if add_deploy_dir and state.deploy_dir:
         dest = os_path.join(state.deploy_dir, dest)
 
@@ -597,7 +619,7 @@ def get(
         if not os_path.exists(local_pathname):
             makedirs(local_pathname)
 
-    remote_file = host.fact.file(src)
+    remote_file = host.get_fact(File, path=src)
 
     # No remote file, so assume exists and download it "blind"
     if not remote_file or force:
@@ -610,7 +632,7 @@ def get(
     # Remote file exists - check if it matches our local
     else:
         local_sum = get_file_sha1(dest)
-        remote_sum = host.fact.sha1_file(src)
+        remote_sum = host.get_fact(Sha1File, path=src)
 
         # Check sha1sum, upload if needed
         if local_sum != remote_sum:
@@ -662,8 +684,6 @@ def put(
         )
     '''
 
-    dest = escape_unix_path(dest)
-
     # Upload IO objects as-is
     if hasattr(src, 'read'):
         local_file = src
@@ -680,7 +700,7 @@ def put(
             raise IOError('No such file: {0}'.format(local_file))
 
     mode = ensure_mode_int(mode)
-    remote_file = host.fact.file(dest)
+    remote_file = host.get_fact(File, path=dest)
 
     if create_remote_dir:
         yield _create_remote_dir(state, host, dest, user, group)
@@ -698,7 +718,7 @@ def put(
     # File exists, check sum and check user/group/mode if supplied
     else:
         local_sum = get_file_sha1(src)
-        remote_sum = host.fact.sha1_file(dest)
+        remote_sum = host.get_fact(Sha1File, path=dest)
 
         # Check sha1sum, upload if needed
         if local_sum != remote_sum:
@@ -811,8 +831,6 @@ def template(
         {% endfor %}
     '''
 
-    dest = escape_unix_path(dest)
-
     if state.deploy_dir:
         src = os_path.join(state.deploy_dir, src)
 
@@ -929,19 +947,18 @@ def link(
     if present and not target:
         raise OperationError('If present is True target must be provided')
 
-    path = escape_unix_path(path)
-    info = host.fact.link(path)
+    info = host.get_fact(Link, path=path)
 
     # Not a link?
     if info is False:
         raise OperationError('{0} exists and is not a link'.format(path))
 
-    add_cmd = 'ln{0} {1} {2}'.format(
-        ' -s' if symbolic else '',
-        target, path,
-    )
+    add_args = ['ln']
+    if symbolic:
+        add_args.append('-s')
 
-    remove_cmd = 'rm -f {0}'.format(path)
+    add_cmd = StringCommand(' '.join(add_args), QuoteString(target), QuoteString(path))
+    remove_cmd = StringCommand('rm', '-f', QuoteString(path))
 
     # No link and we want it
     if not assume_present and info is None and present:
@@ -953,22 +970,22 @@ def link(
         if user or group:
             yield chown(path, user, group, dereference=False)
 
-        host.fact._create(
-            'link',
-            args=(path,),
+        host.create_fact(
+            Link,
+            kwargs={'path': path},
             data={'link_target': target, 'group': group, 'user': user},
         )
 
     # It exists and we don't want it
     elif (assume_present or info) and not present:
         yield remove_cmd
-        host.fact._delete('link', args=(path,))
+        host.delete_fact(Link, kwargs={'path': path})
 
     # Exists and want to ensure it's state
     elif (assume_present or info) and present:
         if assume_present and not info:
             info = {'link_target': None, 'group': None, 'user': None}
-            host.fact._create('link', args=(path,), data=info)
+            host.create_fact(Link, kwargs={'path': path}, data=info)
 
         # If we have an absolute path - prepend to any non-absolute values from the fact
         # and/or the source.
@@ -1055,8 +1072,7 @@ def file(
     _validate_path(path)
 
     mode = ensure_mode_int(mode)
-    path = escape_unix_path(path)
-    info = host.fact.file(path)
+    info = host.get_fact(File, path=path)
 
     # Not a file?!
     if info is False:
@@ -1067,35 +1083,35 @@ def file(
         if create_remote_dir:
             yield _create_remote_dir(state, host, path, user, group)
 
-        yield 'touch {0}'.format(path)
+        yield StringCommand('touch', QuoteString(path))
 
         if mode:
             yield chmod(path, mode)
         if user or group:
             yield chown(path, user, group)
 
-        host.fact._create(
-            'file',
-            args=(path,),
+        host.create_fact(
+            File,
+            kwargs={'path': path},
             data={'mode': mode, 'group': group, 'user': user},
         )
 
     # It exists and we don't want it
     elif (assume_present or info) and not present:
-        yield 'rm -f {0}'.format(path)
-        host.fact._delete('file', args=(path,))
+        yield StringCommand('rm', '-f', QuoteString(path))
+        host.delete_fact(File, kwargs={'path': path})
 
     # It exists & we want to ensure its state
     elif (assume_present or info) and present:
         if assume_present and not info:
             info = {'mode': None, 'group': None, 'user': None}
-            host.fact._create('file', args=(path,), data=info)
+            host.create_fact(File, kwargs={'path': path}, data=info)
 
         changed = False
 
         if touch:
             changed = True
-            yield 'touch {0}'.format(path)
+            yield StringCommand('touch', QuoteString(path))
 
         # Check mode
         if mode and (not info or info['mode'] != mode):
@@ -1171,40 +1187,39 @@ def directory(
     _validate_path(path)
 
     mode = ensure_mode_int(mode)
-    path = escape_unix_path(path)
-    info = host.fact.directory(path)
+    info = host.get_fact(Directory, path=path)
 
     # Not a directory?!
     if info is False:
-        if _no_fail_on_link and host.fact.link(path):
+        if _no_fail_on_link and host.get_fact(Link, path=path):
             host.noop('directory {0} already exists (as a link)'.format(path))
             return
         raise OperationError('{0} exists and is not a directory'.format(path))
 
     # Doesn't exist & we want it
     if not assume_present and info is None and present:
-        yield 'mkdir -p {0}'.format(path)
+        yield StringCommand('mkdir', '-p', QuoteString(path))
         if mode:
             yield chmod(path, mode, recursive=recursive)
         if user or group:
             yield chown(path, user, group, recursive=recursive)
 
-        host.fact._create(
-            'directory',
-            args=(path,),
+        host.create_fact(
+            Directory,
+            kwargs={'path': path},
             data={'mode': mode, 'group': group, 'user': user},
         )
 
     # It exists and we don't want it
     elif (assume_present or info) and not present:
-        yield 'rm -rf {0}'.format(path)
-        host.fact._delete('directory', args=(path,))
+        yield StringCommand('rm', '-rf', QuoteString(path))
+        host.delete_fact(Directory, kwargs={'path': path})
 
     # It exists & we want to ensure its state
     elif (assume_present or info) and present:
         if assume_present and not info:
             info = {'mode': None, 'group': None, 'user': None}
-            host.fact._create('directory', args=(path,), data=info)
+            host.create_fact(Directory, kwargs={'path': path}, data=info)
 
         if _no_check_owner_mode:
             return
